// EndUnit.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ArrayImpl;

/**
 * The EndUnit class instances represent virtual or actuale 
 * places or warehouse were temporary finished sub-recipes (produced internally
 * or obtained as procurements) wait to be used in other recipes via a "p"
 * (procurement) step<br><br>
 *
 * an endUnit has its internal number equal to the concluding code
 * used to finish/identify a temporary finished sub-recipe
 *
 * @author Pietro Terna
 */
public class EndUnit extends SwarmObjectImpl implements PTHistogramPlottable
{

    /** the number identifying the unit which is also its end code */
    public int unitNumber;
    /** the number of layers in the simulation */
    public int totalLayerNumber;
    /** declaring unit sensitive to layers */
    public boolean sensitive=true;
    /** the array of temporary lists, if layers are > 1 and sensitive==true */
    public ArrayImpl temporaryListArray;
    /** the list containing order temporary kept in this virtual (or actual)
     *  place */
    public ListImpl  temporaryList;
    /** the modelSwarm (for future uses) */
    ESFrameModelSwarm eSFrameModelSwarm;


    /**
     * the constructor for EndUnit
     */
    public EndUnit (Zone aZone, ESFrameModelSwarm mo, int tln)
    {
	// Call the constructor for the parent class.
	super(aZone);

	int i;

	// the model (for future use)
	eSFrameModelSwarm = mo;

	// the number of layers
	totalLayerNumber=tln;

	// creating the array of internal lists, with at least one element
	temporaryListArray = new ArrayImpl(getZone(), totalLayerNumber);

	// creating the internal lists to keep the items temporary in hold
	// (one list for each layer; if our unit will be declared to be 
	// insensitive to the layers, then we will have unused lists)
	for (i=0;i<totalLayerNumber;i++){
	    temporaryList = new ListImpl (getZone());
	    temporaryListArray.atOffset$put(i,temporaryList);
	}
    }

    /** set unit number */
    public void setUnitNumber(int un)
    {
	// the unit id number.
	unitNumber=un;
	// announce the unit to the console
	if(StartESFrame.verbose)
	System.out.println("End unit number " + unitNumber + 
			   " has been created.");
     }

    /** get unit number */
    public int getUnitNumber()
    {
	// the unit id number.
	return unitNumber;
    }


    /**
     * Putting an order in the temporaryList of the unit, in a 
     * given layer
     */
    public void setTemporaryList(Order anOrder)
    {
	int layer;

	layer=anOrder.getOrderLayer();
	if(! sensitive)layer=0;

	temporaryList=(ListImpl)temporaryListArray.atOffset(layer);
        temporaryList.addLast(anOrder);
	if(StartESFrame.verbose)
	    System.out.println("The end unit " + unitNumber +
			       " has received the order # "
			       + anOrder.getOrderNumber());
    }

    /**
     * Removing an order from the temporaryList of the unit
     * and returning its address of the order or of its clone if
     * the original order has multiplicity > 1
     */
    public Object removeFromTemporaryList(int cl)
    {
	int chosenLayer;
	Order removedOrder, orderClone;
        
	chosenLayer=cl;
	if(! sensitive)chosenLayer=0;

	temporaryList=(ListImpl) temporaryListArray.atOffset(chosenLayer);

	removedOrder = (Order) temporaryList.removeFirst();
	if(StartESFrame.verbose)
	    System.out.println("The end unit " + unitNumber +
			       " has removed the order # "
			       + removedOrder.getOrderNumber());

	if(removedOrder.getMultiplicity() == 1) return removedOrder;

	else
	    {

		if(StartESFrame.verbose)
		    System.out.println("A clone of the order # " +
		     removedOrder.getOrderNumber() + 
		     " has been created; the original order has now " +
		     "multiplicity " + (removedOrder.getMultiplicity()-1));
		orderClone = (Order) removedOrder.clone();
		removedOrder.setMultiplicity(removedOrder.getMultiplicity()-1);
		temporaryList.addFirst(removedOrder);
		return orderClone;
	    }
    }



    /**
     * return the count of the orders in all the lists, with multiplicity
     */
    public int getAllTemporaryListCount()
    {
	int i, ii, c;

	c = 0;
	for(i=0;i<totalLayerNumber;i++){
	    temporaryList=(ListImpl) temporaryListArray.atOffset(i);	    
	    for (ii=0; ii< temporaryList.getCount();ii++)
		c += ((Order) temporaryList.atOffset(ii)).getMultiplicity();
	}

	return c;
    }

    /**
     * return the count of the orders in a specific list, with multiplicity
     */
    public int getTemporaryListCount(int cl)
    {
	int i, c, chosenLayer;

	chosenLayer=cl;
	if(! sensitive)chosenLayer=0;
	c = 0;

	temporaryList=(ListImpl) temporaryListArray.atOffset(chosenLayer);	    
	for (i=0; i< temporaryList.getCount();i++)
	    c += ((Order) temporaryList.atOffset(i)).getMultiplicity();

	return c;
    }

    /** declaring the end unit insensitive to layers (its number is 
     *  reported as negative in unitData/endUnitList.txt)
     */
    public void setNotSensitiveToLayers(){
	sensitive=false;
	totalLayerNumber=1; // number of layers to be used internally
    }


    /**
     * PTHistogramPlottable interface method: getLabel() 
     */
    public String getLabel(){
	return "" + getUnitNumber();
    }

    /**
     * PTHistogramPlottable interface method: getValueToPlot() 
     */
    public double getValueToPlot(){
	return (double) getAllTemporaryListCount();
    }

}
